/**
  "September 14" CPU emulator.
  Orion Sky Lawlor, olawlor@acm.org, 2005/09/14 (Public Domain)
*/
#include <stdio.h>
#include <stdlib.h>

/**
Quit the emulator.
*/
void quit(const char *why,int value1=0,int value2=0) {
	fprintf(stderr,why,value1,value2);
#ifdef WIN32
	printf("\nPress return to exit...\n");
	getc(stdin);
#endif
	exit(0);
}

/** A tiny CPU Emulator */
class CPU {
public:
	enum {n_regs=16};
	int regs[n_regs]; ///< "registers"-- temporary working space of machine
	enum {n_mem=1024};
	int mem[n_mem]; ///< "memory"-- just program memory for now
	
	int run(void);
	void dump(int id); /* Print out registers */
};

/*************** Main Emulator Run Routine *************/
int CPU::run(void) {
	while (1) {
		int inst=mem[regs[0xf]++]; /* fetch */
		int opcode=(inst>>12)&0xf, i=(inst>>8)&0xf, j=(inst>>4)&0xf, k=(inst>>0)&0xf;
		switch (opcode) { /* Decode */
		case 0x1: regs[i]=(inst&0xff); break; /* "load immediate" */
		case 0xA: regs[i]=regs[j]+regs[k]; break; /* Addition */
		case 0xC: if (regs[j]>=regs[k]) regs[i]++; break; /* Conditional */
		case 0xE: /* Emulator (or OS) call: */
			switch (i) {
			case 0x0: printf("%d%c",regs[k]," \n\t,"[j]); break; /* output */
			case 0x1: 
				printf("Enter a value:\n");  
				scanf("%i",&regs[k]); 
				fgetc(stdin); /* Get rid of terminating newline */
				break; /* output */
			case 0xD: dump(inst&0xff); break; /* Emulator Debug/Dump */
			case 0xE: return regs[k]; /* Emulator Exit */
			default: quit("Unknown OS-trap instruction 0x%X at 0x%X!\n",inst,regs[0xf]-1);
			};
			break;
		default: quit("Unknown instruction 0x%X at 0x%X!\n",inst,regs[0xf]-1);
		};
	}
}

void CPU::dump(int id) {
	if (id==0xff) return; /* 0xEDFF is a disabled dump */
	printf("Dump %02x:\n",id);
	if (id==0)
	for (int m=0;m<n_mem;m++)
		if (mem[m]!=0)
			printf("Memory[%X]=0x%X \n",m,mem[m]);
			
	for (int i=0;i<16;i++)
		if (regs[i]!=0)
			printf("Register[%X]=0x%X  (or %d)\n",i,regs[i],regs[i]);
	printf("\n");
}

int main(int argc,char *argv[]) {
	CPU c={0};
	int initLoc=0x80; /* Location in memory where program is loaded */
	c.regs[0xf]=initLoc; /* Start at first instruction of program */
	
	/* Read program to run from the input file */
	const char *inFile="input.txt"; if (argc>1) inFile=argv[1];
	FILE *f=fopen(inFile,"r"); if (f==NULL) quit("Can't open input file");
	char inLine[1024];
	int loadCount=0;
	while (NULL!=fgets(inLine,1024,f)) {
		int instruction; 
		if (1==sscanf(inLine,"%i",&instruction))
			c.mem[initLoc+loadCount++]=instruction;
	}
	fclose(f);
	
	/* Run the program */
	int v=c.run();
	quit("Program complete.  Return code: %d  (0x%X)\n",v,v);
	return 0;
}
